/*
 * command.c -- command interpreter
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "arith.h"
#include "command.h"
#include "greg.h"
#include "instr.h"
#include "asm.h"
#include "disasm.h"
#include "cpu.h"
#include "mmu.h"
#include "memory.h"
#include "timer.h"
#include "term.h"
#include "disk.h"


#define MAX_TOKENS	10


static bool quit;


static void help(void) {
  cPrintf("valid commands are:\n");
  cPrintf("  +  <num1> <num2>  add and subtract <num1> and <num2>\n");
  cPrintf("  a                 assemble starting at PC\n");
  cPrintf("  a  <addr>         assemble starting at <addr>\n");
  cPrintf("  b                 reset break\n");
  cPrintf("  b  <addr>         set break at <addr>\n");
  cPrintf("  c                 continue execution\n");
  cPrintf("  c  <cnt>          continue execution <cnt> times\n");
  cPrintf("  d                 dump 256 bytes starting at PC\n");
  cPrintf("  d  <addr>         dump 256 bytes starting at <addr>\n");
  cPrintf("  d  <addr> <cnt>   dump <cnt> bytes starting at <addr>\n");
  cPrintf("  i                 initialize hardware\n");
  cPrintf("  u                 unassemble 16 instrs starting at PC\n");
  cPrintf("  u  <addr>         unassemble 16 instrs starting at <addr>\n");
  cPrintf("  u  <addr> <cnt>   unassemble <cnt> instrs starting at <addr>\n");
  cPrintf("  mo                show memory octabyte at PC\n");
  cPrintf("  mo <addr>         show memory octabyte at <addr>\n");
  cPrintf("  mo <addr> <data>  set memory octabyte at <addr> to <data>\n");
  cPrintf("  mt                show memory tetrabyte at PC\n");
  cPrintf("  mt <addr>         show memory tetrabyte at <addr>\n");
  cPrintf("  mt <addr> <data>  set memory tetrabyte at <addr> to <data>\n");
  cPrintf("  mw                show memory wyde at PC\n");
  cPrintf("  mw <addr>         show memory wyde at <addr>\n");
  cPrintf("  mw <addr> <data>  set memory wyde at <addr> to <data>\n");
  cPrintf("  mb                show memory byte at PC\n");
  cPrintf("  mb <addr>         show memory byte at <addr>\n");
  cPrintf("  mb <addr> <data>  set memory byte at <addr> to <data>\n");
  cPrintf("  @                 show PC\n");
  cPrintf("  @  <addr>         set PC to <addr>\n");
  cPrintf("  g                 show available global registers\n");
  cPrintf("  g  <reg>          show global register <reg>\n");
  cPrintf("  g  <reg> <data>   set global register <reg> to <data>\n");
  cPrintf("  l                 show available local registers\n");
  cPrintf("  l  <reg>          show local register <reg>\n");
  cPrintf("  l  <reg> <data>   set local register <reg> to <data>\n");
  cPrintf("  $                 show available dynamic registers\n");
  cPrintf("  $  <reg>          show dynamic register <reg>\n");
  cPrintf("  $  <reg> <data>   set dynamic register <reg> to <data>\n");
  cPrintf("  s                 single-step one instruction\n");
  cPrintf("  s  <cnt>          single-step <cnt> instructions\n");
  cPrintf("  q                 quit simulator\n");
}


static int getHexNumber(char *str, octa *valptr) {
  char *end;
  int n, k;
  tetra val;

  n = strlen(str);
  end = str + n - 1;
  valptr->h = 0;
  valptr->l = 0;
  for (k = 0; k < n; k++) {
    if (*end >= '0' && *end <= '9') {
      val = *end - '0';
    } else
    if (*end >= 'a' && *end <= 'f') {
      val = *end - 'a' + 10;
    } else
    if (*end >= 'A' && *end <= 'F') {
      val = *end - 'A' + 10;
    } else {
      /* number is malformed */
      return 0;
    }
    end--;
    if (k >= 8) {
      valptr->h += val << (4 * k - 32);
    } else {
      valptr->l += val << (4 * k);
    }
  }
  return 1;
}


static int getInteger(char *str, int *valptr) {
  char *end;

  *valptr = strtoul(str, &end, 10);
  return *end == '\0';
}


static void showPC(void) {
  octa pc;
  tetra instr;

  pc = cpuGetPC();
  instr = mmuRead32(pc);
  cPrintf("PC     %08X%08X     [PC]   %08X   %s\n",
          pc.h, pc.l, instr, disasm(instr, pc));
}


static void showBreak(void) {
  octa brk;

  brk = cpuGetBreak();
  cPrintf("Break  ");
  if (cpuTestBreak()) {
    cPrintf("%08X%08X", brk.h, brk.l);
  } else {
    cPrintf("-- not set --");
  }
  cPrintf("\n");
}


static void doArith(char *tokens[], int n) {
  octa num1, num2, num3, num4;

  if (n == 3) {
    if (!getHexNumber(tokens[1], &num1)) {
      cPrintf("illegal first number\n");
      return;
    }
    if (!getHexNumber(tokens[2], &num2)) {
      cPrintf("illegal second number\n");
      return;
    }
    num3 = addOcta(num1, num2);
    num4 = subOcta(num1, num2);
    cPrintf("add = %08X%08X, sub = %08X%08X\n",
            num3.h, num3.l, num4.h, num4.l);
  } else {
    help();
  }
}


static void doAssemble(char *tokens[], int n) {
  octa addr;
  char prompt[30];
  char *line;
  tetra instr;
  char *msg;

  if (n == 1) {
    addr = cpuGetPC();
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    addr.l &= ~0x00000003;
  } else {
    help();
    return;
  }
  while (1) {
    sprintf(prompt, "ASM @ %08X%08X: ", addr.h, addr.l);
    line = cGetLine(prompt);
    if (*line == '\0' || *line == '\n') {
      break;
    }
    cAddHist(line);
    msg = asmInstr(line, addr, &instr);
    if (msg != NULL) {
      cPrintf("%s\n", msg);
    } else {
      mmuWrite32(addr, instr);
      addr = incrOcta(addr, 4);
    }
  }
}


static void doBreak(char *tokens[], int n) {
  octa addr;

  if (n == 1) {
    cpuResetBreak();
    showBreak();
    showPC();
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal break address\n");
      return;
    }
    cpuSetBreak(addr);
    showBreak();
    showPC();
  } else {
    help();
  }
}


static void doContinue(char *tokens[], int n) {
  octa tmp;
  tetra count, i;
  octa brk;

  if (n == 1) {
    count = 1;
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &tmp) || tmp.h != 0 || tmp.l == 0) {
      cPrintf("illegal continue count\n");
      return;
    }
    count = tmp.l;
  } else {
    help();
    return;
  }
  for (i = 0; i < count; i++) {
    cpuContinue();
  }
  brk = cpuGetBreak();
  cPrintf("\nBreak at %08X%08X\n", brk.h, brk.l);
}


static void doDump(char *tokens[], int n) {
  octa addr, count;
  octa lo, hi, curr;
  int lines, i, j;
  byte c;
  octa tmp;

  if (n == 1) {
    addr = cpuGetPC();
    count.h = 0;
    count.l = 16 * 16;
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal dump address\n");
      return;
    }
    count.h = 0;
    count.l = 16 * 16;
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal dump address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &count)) {
      cPrintf("illegal dump count\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  lo.h = addr.h;
  lo.l = addr.l & ~0x0000000F;
  hi = incrOcta(addOcta(addr, count), -1);
  if (hi.h < lo.h || (hi.h == lo.h && hi.l < lo.l)) {
    /* wrap-around */
    hi.h = 0xFFFFFFFF;
    hi.l = 0xFFFFFFFF;
  }
  lines = incrOcta(subOcta(hi, lo), 16).l >> 4;
  curr = lo;
  for (i = 0; i < lines; i++) {
    cPrintf("%08X%08X:  ", curr.h, curr.l);
    for (j = 0; j < 16; j++) {
      tmp = incrOcta(curr, j);
      if (tmp.h < addr.h || (tmp.h == addr.h && tmp.l < addr.l) ||
          tmp.h > hi.h || (tmp.h == hi.h && tmp.l > hi.l)) {
        cPrintf("  ");
      } else {
        c = mmuRead8(tmp);
        cPrintf("%02X", c);
      }
      if ((j & 3) == 3) {
        cPrintf(" ");
      }
    }
    cPrintf("  ");
    for (j = 0; j < 16; j++) {
      tmp = incrOcta(curr, j);
      if (tmp.h < addr.h || (tmp.h == addr.h && tmp.l < addr.l) ||
          tmp.h > hi.h || (tmp.h == hi.h && tmp.l > hi.l)) {
        cPrintf(" ");
      } else {
        c = mmuRead8(tmp);
        if (c >= 32 && c <= 126) {
          cPrintf("%c", c);
        } else {
          cPrintf(".");
        }
      }
    }
    cPrintf("\n");
    curr = incrOcta(curr, 16);
  }
}


static void doInit(char *tokens[], int n) {
  if (n == 1) {
    timerReset();
    termReset();
    diskReset();
    memoryReset();
    mmuReset();
    cpuReset();
  } else {
    help();
  }
}


static void doUnassemble(char *tokens[], int n) {
  octa addr, count;
  int i;
  tetra instr;

  if (n == 1) {
    addr = cpuGetPC();
    count.h = 0;
    count.l = 16;
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal list address\n");
      return;
    }
    count.h = 0;
    count.l = 16;
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal list address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &count)) {
      cPrintf("illegal list count\n");
      return;
    }
    if (count.h == 0 && count.l == 0) {
      return;
    }
  } else {
    help();
    return;
  }
  addr.l &= ~0x00000003;
  for (i = 0; i < count.l; i++) {
    instr = mmuRead32(addr);
    cPrintf("%08X%08X:  %08X    %s\n",
            addr.h, addr.l, instr, disasm(instr, addr));
    addr = incrOcta(addr, 4);
    if (addr.h == 0 && addr.l == 0) {
      /* wrap-around */
      break;
    }
  }
}


static void doMemoryOcta(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  octa newData;
  octa data;

  if (n == 1) {
    addr = cpuGetPC();
    data = mmuRead64(addr);
    cPrintf("%08X%08X:  %08X%08X\n",
            addr.h, addr.l, data.h, data.l);
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    data = mmuRead64(addr);
    cPrintf("%08X%08X:  %08X%08X\n",
            addr.h, addr.l, data.h, data.l);
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData;
    data = mmuRead64(addr);
    mmuWrite64(addr, newData);
    cPrintf("%08X%08X:  %08X%08X  %08X%08X\n",
            addr.h, addr.l, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doMemoryTetra(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  tetra newData;
  tetra data;

  if (n == 1) {
    addr = cpuGetPC();
    data = mmuRead32(addr);
    cPrintf("%08X%08X:  %08X\n",
            addr.h, addr.l, data);
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    data = mmuRead32(addr);
    cPrintf("%08X%08X:  %08X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = tmpData.l;
    data = mmuRead32(addr);
    mmuWrite32(addr, newData);
    cPrintf("%08X%08X:  %08X  %08X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryWyde(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  wyde newData;
  wyde data;

  if (n == 1) {
    addr = cpuGetPC();
    data = mmuRead16(addr);
    cPrintf("%08X%08X:  %04X\n",
            addr.h, addr.l, data);
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    data = mmuRead16(addr);
    cPrintf("%08X%08X:  %04X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (wyde) tmpData.l;
    data = mmuRead16(addr);
    mmuWrite16(addr, newData);
    cPrintf("%08X%08X:  %04X  %04X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doMemoryByte(char *tokens[], int n) {
  octa addr;
  octa tmpData;
  byte newData;
  byte data;

  if (n == 1) {
    addr = cpuGetPC();
    data = mmuRead8(addr);
    cPrintf("%08X%08X:  %02X\n",
            addr.h, addr.l, data);
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    data = mmuRead8(addr);
    cPrintf("%08X%08X:  %02X\n",
            addr.h, addr.l, data);
  } else if (n == 3) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    if (!getHexNumber(tokens[2], &tmpData)) {
      cPrintf("illegal data\n");
      return;
    }
    newData = (byte) tmpData.l;
    data = mmuRead8(addr);
    mmuWrite8(addr, newData);
    cPrintf("%08X%08X:  %02X  %02X\n",
            addr.h, addr.l, data, newData);
  } else {
    help();
  }
}


static void doPC(char *tokens[], int n) {
  octa addr;

  if (n == 1) {
    showBreak();
    showPC();
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &addr)) {
      cPrintf("illegal address\n");
      return;
    }
    cpuSetPC(addr);
    showBreak();
    showPC();
  } else {
    help();
  }
}


static void doGReg(char *tokens[], int n) {
  int reg;
  Greg *rp;
  octa data;
  octa newData;

  if (n == 1) {
    cPrintf("Global registers %d..%d present\n", 0, GR_NUM - 1);
    for (reg = 0; reg < 32; reg++) {
      cPrintf("g[%2d]  %-3s  %-27s",
              reg, gregCodeTbl[reg]->name, gregCodeTbl[reg]->description);
      if (reg % 2 == 1) {
        cPrintf("\n");
      }
    }
  } else if (n == 2) {
    if (!getInteger(tokens[1], &reg)) {
      rp = lookupGreg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (reg < 0 || reg > GR_NUM - 1) {
        cPrintf("illegal register number\n");
        return;
      }
    }
    data = cpuGetGR(reg);
    if (reg < 32) {
      cPrintf("%-6s  %08X%08X\n", gregCodeTbl[reg]->name, data.h, data.l);
    } else {
      cPrintf("g[%3d]  %08X%08X\n", reg, data.h, data.l);
    }
  } else if (n == 3) {
    if (!getInteger(tokens[1], &reg)) {
      rp = lookupGreg(tokens[1]);
      if (rp == NULL) {
        cPrintf("illegal register name\n");
        return;
      }
      reg = rp->code;
    } else {
      if (reg < 0 || reg > GR_NUM - 1) {
        cPrintf("illegal register number\n");
        return;
      }
    }
    if (!getHexNumber(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetGR(reg);
    cpuSetGR(reg, newData);
    if (reg < 32) {
      cPrintf("%-6s  %08X%08X  %08X%08X\n",
              gregCodeTbl[reg]->name, data.h, data.l, newData.h, newData.l);
    } else {
      cPrintf("g[%3d]  %08X%08X  %08X%08X\n",
              reg, data.h, data.l, newData.h, newData.l);
    }
  } else {
    help();
  }
}


static void doLReg(char *tokens[], int n) {
  int reg;
  octa data;
  octa newData;

  if (n == 1) {
    cPrintf("Local registers %d..%d present\n", 0, LR_NUM - 1);
  } else if (n == 2) {
    if (!getInteger(tokens[1], &reg) || reg < 0 || reg > LR_NUM - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    data = cpuGetLR(reg);
    cPrintf("l[%3d]  %08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getInteger(tokens[1], &reg) || reg < 0 || reg > LR_NUM - 1) {
      cPrintf("illegal register number\n");
      return;
    }
    if (!getHexNumber(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetLR(reg);
    cpuSetLR(reg, newData);
    cPrintf("l[%3d]  %08X%08X  %08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


static void doDReg(char *tokens[], int n) {
  int L, G;
  int reg;
  octa data;
  octa newData;

  if (n == 1) {
    L = cpuGetGR(rL).l;
    G = cpuGetGR(rG).l;
    if (0 <= L - 1) {
      cPrintf("Dynamic registers %d..%d are local\n", 0, L - 1);
    }
    if (L <= G - 1) {
      cPrintf("Dynamic registers %d..%d are marginal\n", L, G - 1);
    }
    if (G <= 255) {
      cPrintf("Dynamic registers %d..%d are global\n", G, 255);
    }
  } else if (n == 2) {
    if (!getInteger(tokens[1], &reg) || reg < 0 || reg > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    data = cpuGetDR(reg);
    cPrintf("$[%3d]  %08X%08X\n", reg, data.h, data.l);
  } else if (n == 3) {
    if (!getInteger(tokens[1], &reg) || reg < 0 || reg > 255) {
      cPrintf("illegal register number\n");
      return;
    }
    if (!getHexNumber(tokens[2], &newData)) {
      cPrintf("illegal data\n");
      return;
    }
    data = cpuGetDR(reg);
    cpuSetDR(reg, newData);
    cPrintf("$[%3d]  %08X%08X  %08X%08X\n",
            reg, data.h, data.l, newData.h, newData.l);
  } else {
    help();
  }
}


//static void doCPURegister(char *tokens[], int n) {
//  int reg;
//  int32 data;
//  int32 newData;
//
//  if (n == 1) {
//    showCPURegs();
//    showBreak();
//    showPC();
//  } else if (n == 2) {
//    if (strcmp(tokens[1], "pc") == 0) {
//      data = cpuGetPC();
//      cPrintf("PC   %08X\n", data);
//      return;
//    }
//    if (tokens[1][0] != 'r') {
//      cPrintf("illegal register name\n");
//      return;
//    }
//    if (!getInteger(&tokens[1][1], &reg) ||
//        reg < 0 || reg > NUM_REGS - 1) {
//      cPrintf("illegal register number\n");
//      return;
//    }
//    data = cpuGetGR(reg);
//    cPrintf("r%-2d  %08X\n", reg, data);
//  } else if (n == 3) {
//    if (!getHexNumber(tokens[2], &newData)) {
//      cPrintf("illegal data\n");
//      return;
//    }
//    if (strcmp(tokens[1], "pc") == 0) {
//      data = cpuGetPC();
//      cpuSetPC(newData);
//      cPrintf("PC   %08X  %08X\n", data, newData);
//      return;
//    }
//    if (tokens[1][0] != 'r') {
//      cPrintf("illegal register name\n");
//      return;
//    }
//    if (!getInteger(&tokens[1][1], &reg) ||
//        reg < 0 || reg > NUM_REGS - 1) {
//      cPrintf("illegal register number\n");
//      return;
//    }
//    data = cpuGetGR(reg);
//    cpuSetGR(reg, newData);
//    cPrintf("r%-2d  %08X  %08X\n", reg, data, newData);
//  } else {
//    help();
//  }
//}


static void doStep(char *tokens[], int n) {
  octa tmp;
  tetra count, i;

  if (n == 1) {
    count = 1;
  } else if (n == 2) {
    if (!getHexNumber(tokens[1], &tmp) || tmp.h != 0 || tmp.l == 0) {
      cPrintf("illegal step count\n");
      return;
    }
    count = tmp.l;
  } else {
    help();
    return;
  }
  for (i = 0; i < count; i++) {
    cpuStep();
  }
  showPC();
}


static void doQuit(char *tokens[], int n) {
  if (n == 1) {
    quit = TRUE;
  } else {
    help();
  }
}


typedef struct {
  char *name;
  void (*cmdProc)(char *tokens[], int n);
} Command;

Command commands[] = {
  { "+",  doArith },
  { "a",  doAssemble },
  { "b",  doBreak },
  { "c",  doContinue },
  { "d",  doDump },
  { "i",  doInit },
  { "u",  doUnassemble },
  { "mo", doMemoryOcta },
  { "mt", doMemoryTetra },
  { "mw", doMemoryWyde },
  { "mb", doMemoryByte },
  { "@",  doPC },
  { "g",  doGReg },
  { "l",  doLReg },
  { "$",  doDReg },
  { "s",  doStep },
  { "q",  doQuit }
};


bool doCommand(char *line) {
  char *tokens[MAX_TOKENS];
  int n;
  char *p;
  int i;

  n = 0;
  p = strtok(line, " \t\n");
  while (p != NULL) {
    if (n == MAX_TOKENS) {
      cPrintf("too many tokens on line\n");
      return 0;
    }
    tokens[n++] = p;
    p = strtok(NULL, " \t\n");
  }
  if (n == 0) {
    return FALSE;
  }
  quit = FALSE;
  for (i = 0; i < sizeof(commands)/sizeof(commands[0]); i++) {
    if (strcmp(commands[i].name, tokens[0]) == 0) {
      (*commands[i].cmdProc)(tokens, n);
      return quit;
    }
  }
  help();
  return FALSE;
}


static void handleInteractiveException(int exception) {
  char *what;

  what = exceptionToString(exception);
  cPrintf("\n");
  cPrintf("NOTE: %s occurred while executing the command.\n", what);
  cPrintf("      This event will not alter the state of the CPU.\n");
}


bool execCommand(char *line) {
  jmp_buf myEnvironment;
  int exception;
  bool quit;

  exception = setjmp(myEnvironment);
  if (exception != NO_EXCEPTION) {
    /* exception thrown */
    handleInteractiveException(exception);
    popEnvironment();
    return FALSE;
  }
  pushEnvironment(&myEnvironment);
  quit = doCommand(line);
  popEnvironment();
  return quit;
}
