/*
 * cpu.c -- CPU simulation
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "curslib.h"
#include "error.h"
#include "arith.h"
#include "greg.h"
#include "instr.h"
#include "cpu.h"
#include "mmu.h"
#include "timer.h"

#include "abstime.h"


static octa pc;			/* program counter */
static octa gr[GR_NUM];		/* global registers */
static octa lr[LR_NUM];		/* local registers */

static int G;			/* easily accessible copy of rG */
static int L;			/* easily accessible copy of rL */
static int O;			/* congruent to rO >> 3 modulo LR_NUM */
static int S;			/* congruent to rS >> 3 modulo LR_NUM */

static bool breakSet;		/* is break active? */
static octa breakAddr;		/* where to break */

static int instrCount;		/* count to tick the timer */

static bool debugStack = FALSE;


/*------------------------------------------------------------*/


static void handleRuntimeException(int exception) {
  cPrintf("**** runtime exception %d thrown ****\n", exception);
}


/*------------------------------------------------------------*/


static void stackStore(void) {
  int gamma;

  gamma = S & LR_MASK;
  mmuWrite64(gr[rS], lr[gamma]);
  if (debugStack) {
    cPrintf("\nSTACK store: M8[#%08X%08X]=l[%d](#%08X%08X), rS+=8\n",
            gr[rS].h, gr[rS].l, gamma, lr[gamma].h, lr[gamma].l);
  }
  gr[rS] = incrOcta(gr[rS], 8);
  S++;
}


static void stackLoad(void) {
  int gamma;

  S--;
  gr[rS] = incrOcta(gr[rS], -8);
  gamma = S & LR_MASK;
  lr[gamma] = mmuRead64(gr[rS]);
  if (debugStack) {
    cPrintf("\nSTACK load:  rS-=8, l[%d]=M8[#%08X%08X](#%08X%08X)\n",
            gamma, gr[rS].h, gr[rS].l, lr[gamma].h, lr[gamma].l);
  }
}


/*------------------------------------------------------------*/


static octa readDR(int reg) {
  if (reg >= G) {
    /* global register */
    return gr[reg];
  }
  if (reg < L) {
    /* local register */
    return lr[(O + reg) & LR_MASK];
  }
  /* marginal register */
  return zero;
}


static void writeDR(int reg, octa data) {
  int beta;

  if (reg >= G) {
    /* global register */
    gr[reg] = data;
  } else {
    /* marginal or local register */
    while (reg >= L) {
      /* increase rL */
      beta = (O + L) & LR_MASK;
      lr[beta] = zero;
      L++;
      gr[rL].l = L;
      if (debugStack) {
        cPrintf("\nSTACK incr:  beta=%d, rL=%d\n",
                (beta + 1) & LR_MASK, L);
      }
      if (((S - O - L) & LR_MASK) == 0) {
        stackStore();
      }
    }
    /* reg is now a local register */
    lr[(O + reg) & LR_MASK] = data;
  }
}


/*------------------------------------------------------------*/


static void push(int reg) {
  octa numRegs;

  if (reg >= G) {
    reg = L;
  }
  numRegs.h = 0;
  numRegs.l = reg;
  writeDR(reg, numRegs);
  O += reg + 1;
  gr[rO] = incrOcta(gr[rO], (reg + 1) << 3);
  L -= reg + 1;
  gr[rL].l = L;
  if (debugStack) {
    cPrintf("\nSTACK push:  l[%d]=%d, alpha=%d, rL=%d\n",
            (O - 1) & LR_MASK, reg, O & LR_MASK, L);
  }
}


static void pop(int reg) {
  octa holeData;
  int numRegs;

  if (reg != 0 && reg <= L) {
    holeData = lr[(O + reg - 1) & LR_MASK];
  }
  if (gr[rS].l == gr[rO].l) {
    stackLoad();
  }
  numRegs = lr[(O - 1) & LR_MASK].l & 0xFF;
  while ((tetra) (O - S) <= (tetra) numRegs) {
    stackLoad();
  }
  L = numRegs + (reg <= L ? reg : L + 1);
  if (L > G) {
    L = G;
  }
  if (L > numRegs) {
    lr[(O - 1) & LR_MASK] = holeData;
  }
  O -= numRegs + 1;
  gr[rO] = incrOcta(gr[rO], -((numRegs + 1) << 3));
  gr[rL].l = L;
  if (debugStack) {
    if (L > numRegs) {
      cPrintf("\nSTACK pop:   l[%d]=#%08X%08X, alpha=%d, rL=%d\n",
              (O + numRegs) & LR_MASK, holeData, O & LR_MASK, L);
    } else {
      cPrintf("\nSTACK pop:   alpha=%d, rL=%d\n",
              O & LR_MASK, L);
    }
  }
}


/*------------------------------------------------------------*/


static octa readSpecial(int reg) {
  /* every special register is readable */
  return gr[reg];
}


static void writeSpecial(int reg, octa data) {
  int i;

  /* writing to certain special registers is restricted */
  if (reg >= 8 && reg <= 11) {
    /* nobody is allowed to write to these registers: */
    /* rC, rN, rO, rS */
    /* !!!!! signal illegal instruction exception !!!!! */
  } else
  if (reg >= 12 && reg <= 18) {
    /* only the operating system may write to these registers: */
    /* rI, rT, rTT, rK, rQ, rU, rV */
    if ((gr[rK].h & k_BIT) != 0) {
      /* !!!!! signal privileged operation exception !!!!! */
    } else {
      gr[reg] = data;
    }
  } else
  if (reg == rA) {
    /* only the lower 18 bits may change */
    if (data.h != 0 || data.l >= 0x40000) {
      /* !!!!! signal illegal instruction exception !!!!! */
    } else {
      gr[rA] = data;
    }
  } else
  if (reg == rL) {
    /* set rL to min(data,rL) */
    if (data.h == 0 && data.l < L) {
      L = data.l;
      gr[rL].l = L;
    }
  } else
  if (reg == rG) {
    /* value must be in range 32..255 and not less than rL */
    if (data.h != 0 || data.l > 255 || data.l < L || data.l < 32) {
      /* !!!!! signal illegal instruction exception !!!!! */
    } else {
      for (i = data.l; i < G; i++) {
        gr[i] = zero;
      }
      G = data.l;
      gr[rG].l = G;
    }
  } else {
    /* the remaining registers may be written to without restrictions */
    gr[reg] = data;
  }
}


/*------------------------------------------------------------*/


static void ill(tetra instr) {
  throwException(INSTRUCTION_EXCEPTION);
}


/*------------------------------------------------------------*/

/* 00 - 0F */


static void TRAP(tetra instr) {
  /* !!!!! not implemented yet !!!!! */
  ill(instr);
}


/*------------------------------------------------------------*/

/* 10 - 1F */


static void MUL(tetra instr) {
  /* multiply */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = smulOcta(op1, op2);
  if (arithOvrflw) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void MULI(tetra instr) {
  /* multiply immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = smulOcta(op1, op2);
  if (arithOvrflw) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void MULU(tetra instr) {
  /* multiply unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = umulOcta(op1, op2);
  gr[rH] = arithAuxOut;
  writeDR(x, res);
}


static void MULUI(tetra instr) {
  /* multiply unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = umulOcta(op1, op2);
  gr[rH] = arithAuxOut;
  writeDR(x, res);
}


static void DIV(tetra instr) {
  /* divide */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = sdivOcta(op1, op2);
  gr[rR] = arithAuxOut;
  if (arithOvrflw) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void DIVI(tetra instr) {
  /* divide immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = sdivOcta(op1, op2);
  gr[rR] = arithAuxOut;
  if (arithOvrflw) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void DIVU(tetra instr) {
  /* divide unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = udivOcta(gr[rD], op1, op2);
  gr[rR] = arithAuxOut;
  writeDR(x, res);
}


static void DIVUI(tetra instr) {
  /* divide unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = udivOcta(gr[rD], op1, op2);
  gr[rR] = arithAuxOut;
  writeDR(x, res);
}


/*------------------------------------------------------------*/

/* 20 - 2F */


static void ADD(tetra instr) {
  /* add */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(op1, op2);
  if (((op1.h ^ op2.h) & MSB(32)) == 0 &&
      ((op1.h ^ res.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void ADDI(tetra instr) {
  /* add immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(op1, op2);
  if (((op1.h ^ op2.h) & MSB(32)) == 0 &&
      ((op1.h ^ res.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void ADDU(tetra instr) {
  /* add unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(op1, op2);
  writeDR(x, res);
}


static void ADDUI(tetra instr) {
  /* add unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(op1, op2);
  writeDR(x, res);
}


static void SUB(tetra instr) {
  /* subtract */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = subOcta(op1, op2);
  if (((res.h ^ op2.h) & MSB(32)) == 0 &&
      ((res.h ^ op1.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void SUBI(tetra instr) {
  /* subtract immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = subOcta(op1, op2);
  if (((res.h ^ op2.h) & MSB(32)) == 0 &&
      ((res.h ^ op1.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void SUBU(tetra instr) {
  /* subtract unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = subOcta(op1, op2);
  writeDR(x, res);
}


static void SUBUI(tetra instr) {
  /* subtract unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = subOcta(op1, op2);
  writeDR(x, res);
}


static void ADDU2(tetra instr) {
  /* times 2 and add unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(shlOcta(op1, 1), op2);
  writeDR(x, res);
}


static void ADDUI2(tetra instr) {
  /* times 2 and add unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(shlOcta(op1, 1), op2);
  writeDR(x, res);
}


static void ADDU4(tetra instr) {
  /* times 4 and add unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(shlOcta(op1, 2), op2);
  writeDR(x, res);
}


static void ADDUI4(tetra instr) {
  /* times 4 and add unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(shlOcta(op1, 2), op2);
  writeDR(x, res);
}


static void ADDU8(tetra instr) {
  /* times 8 and add unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(shlOcta(op1, 3), op2);
  writeDR(x, res);
}


static void ADDUI8(tetra instr) {
  /* times 8 and add unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(shlOcta(op1, 3), op2);
  writeDR(x, res);
}


static void ADDU16(tetra instr) {
  /* times 16 and add unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = addOcta(shlOcta(op1, 4), op2);
  writeDR(x, res);
}


static void ADDUI16(tetra instr) {
  /* times 16 and add unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = addOcta(shlOcta(op1, 4), op2);
  writeDR(x, res);
}


/*------------------------------------------------------------*/

/* 30 - 3F */


static void CMP(tetra instr) {
  /* compare */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if ((op1.h & MSB(32)) > (op2.h & MSB(32))) {
    res = negone;
  } else
  if ((op1.h & MSB(32)) < (op2.h & MSB(32))) {
    res = one;
  } else
  if (op1.h < op2.h) {
    res = negone;
  } else
  if (op1.h > op2.h) {
    res = one;
  } else
  if (op1.l < op2.l) {
    res = negone;
  } else
  if (op1.l > op2.l) {
    res = one;
  } else {
    res = zero;
  }
  writeDR(x, res);
}


static void CMPI(tetra instr) {
  /* compare immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if ((op1.h & MSB(32)) > (op2.h & MSB(32))) {
    res = negone;
  } else
  if ((op1.h & MSB(32)) < (op2.h & MSB(32))) {
    res = one;
  } else
  if (op1.h < op2.h) {
    res = negone;
  } else
  if (op1.h > op2.h) {
    res = one;
  } else
  if (op1.l < op2.l) {
    res = negone;
  } else
  if (op1.l > op2.l) {
    res = one;
  } else {
    res = zero;
  }
  writeDR(x, res);
}


static void CMPU(tetra instr) {
  /* compare unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if (op1.h < op2.h) {
    res = negone;
  } else
  if (op1.h > op2.h) {
    res = one;
  } else
  if (op1.l < op2.l) {
    res = negone;
  } else
  if (op1.l > op2.l) {
    res = one;
  } else {
    res = zero;
  }
  writeDR(x, res);
}


static void CMPUI(tetra instr) {
  /* compare unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if (op1.h < op2.h) {
    res = negone;
  } else
  if (op1.h > op2.h) {
    res = one;
  } else
  if (op1.l < op2.l) {
    res = negone;
  } else
  if (op1.l > op2.l) {
    res = one;
  } else {
    res = zero;
  }
  writeDR(x, res);
}


static void NEG(tetra instr) {
  /* negate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1.h = 0;
  op1.l = ZEXT(y, 8);
  op2 = readDR(z);
  res = subOcta(op1, op2);
  if (((res.h ^ op2.h) & MSB(32)) == 0 &&
      ((res.h ^ op1.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void NEGI(tetra instr) {
  /* negate immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1.h = 0;
  op1.l = ZEXT(y, 8);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = subOcta(op1, op2);
  if (((res.h ^ op2.h) & MSB(32)) == 0 &&
      ((res.h ^ op1.h) & MSB(32)) != 0) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void NEGU(tetra instr) {
  /* negate unsigned */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1.h = 0;
  op1.l = ZEXT(y, 8);
  op2 = readDR(z);
  res = subOcta(op1, op2);
  writeDR(x, res);
}


static void NEGUI(tetra instr) {
  /* negate unsigned immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1.h = 0;
  op1.l = ZEXT(y, 8);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = subOcta(op1, op2);
  writeDR(x, res);
}


static void SL(tetra instr) {
  /* shift left */
  int x, y, z;
  octa op1, op2, res;
  int n;
  octa tmp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shlOcta(op1, n);
  tmp = sarOcta(res, n);
  if (tmp.h != op1.h || tmp.l != op1.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void SLI(tetra instr) {
  /* shift left immediate */
  int x, y, z;
  octa op1, op2, res;
  int n;
  octa tmp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shlOcta(op1, n);
  tmp = sarOcta(res, n);
  if (tmp.h != op1.h || tmp.l != op1.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  writeDR(x, res);
}


static void SLU(tetra instr) {
  /* shift left unsigned */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shlOcta(op1, n);
  writeDR(x, res);
}


static void SLUI(tetra instr) {
  /* shift left unsigned immediate */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shlOcta(op1, n);
  writeDR(x, res);
}


static void SR(tetra instr) {
  /* shift right */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = sarOcta(op1, n);
  writeDR(x, res);
}


static void SRI(tetra instr) {
  /* shift right immediate */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = sarOcta(op1, n);
  writeDR(x, res);
}


static void SRU(tetra instr) {
  /* shift right unsigned */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shrOcta(op1, n);
  writeDR(x, res);
}


static void SRUI(tetra instr) {
  /* shift right unsigned immediate */
  int x, y, z;
  octa op1, op2, res;
  int n;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  if (op2.h != 0 || op2.l >= 64) {
    n = 64;
  } else {
    n = op2.l;
  }
  res = shrOcta(op1, n);
  writeDR(x, res);
}


/*------------------------------------------------------------*/

/* 40 - 4F */


static void BN(tetra instr) {
  /* branch if negative */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNB(tetra instr) {
  /* branch if negative back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BZ(tetra instr) {
  /* branch if zero */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h == 0 && data.l == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BZB(tetra instr) {
  /* branch if zero back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h == 0 && data.l == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BP(tetra instr) {
  /* branch if positive */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BPB(tetra instr) {
  /* branch if positive back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BOD(tetra instr) {
  /* branch if odd */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BODB(tetra instr) {
  /* branch if odd back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNN(tetra instr) {
  /* branch if nonnegative */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNNB(tetra instr) {
  /* branch if nonnegative back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNZ(tetra instr) {
  /* branch if nonzero */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h != 0 || data.l != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNZB(tetra instr) {
  /* branch if nonzero back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h != 0 || data.l != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNP(tetra instr) {
  /* branch if nonpositive */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BNPB(tetra instr) {
  /* branch if nonpositive back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BEV(tetra instr) {
  /* branch if even */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void BEVB(tetra instr) {
  /* branch if even back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


/*------------------------------------------------------------*/

/* 50 - 5F */


static void PBN(tetra instr) {
  /* probable branch if negative */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNB(tetra instr) {
  /* probable branch if negative back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBZ(tetra instr) {
  /* probable branch if zero */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h == 0 && data.l == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBZB(tetra instr) {
  /* probable branch if zero back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h == 0 && data.l == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBP(tetra instr) {
  /* probable branch if positive */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBPB(tetra instr) {
  /* probable branch if positive back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBOD(tetra instr) {
  /* probable branch if odd */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBODB(tetra instr) {
  /* probable branch if odd back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNN(tetra instr) {
  /* probable branch if nonnegative */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNNB(tetra instr) {
  /* probable branch if nonnegative back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNZ(tetra instr) {
  /* probable branch if nonzero */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h != 0 || data.l != 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNZB(tetra instr) {
  /* probable branch if nonzero back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if (data.h != 0 || data.l != 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNP(tetra instr) {
  /* probable branch if nonpositive */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBNPB(tetra instr) {
  /* probable branch if nonpositive back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBEV(tetra instr) {
  /* probable branch if even */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) == 0) {
    offset = LIT16(instr) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


static void PBEVB(tetra instr) {
  /* probable branch if even back */
  int x;
  octa data;
  int offset;

  x = DST(instr);
  data = readDR(x);
  if ((data.l & 1) == 0) {
    offset = (LIT16(instr) - 0x10000) << 2;
    pc = incrOcta(pc, offset - 4);
  }
}


/*------------------------------------------------------------*/

/* 60 - 6F */


static void CSN(tetra instr) {
  /* conditionally set if negative */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSNI(tetra instr) {
  /* conditionally set if negative immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSZ(tetra instr) {
  /* conditionally set if zero */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h == 0 && data.l == 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSZI(tetra instr) {
  /* conditionally set if zero immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h == 0 && data.l == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSP(tetra instr) {
  /* conditionally set if positive */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSPI(tetra instr) {
  /* conditionally set if positive immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSOD(tetra instr) {
  /* conditionally set if odd */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) != 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSODI(tetra instr) {
  /* conditionally set if odd immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSNN(tetra instr) {
  /* conditionally set if nonnegative */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSNNI(tetra instr) {
  /* conditionally set if nonnegative immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSNZ(tetra instr) {
  /* conditionally set if nonzero */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h != 0 || data.l != 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSNZI(tetra instr) {
  /* conditionally set if nonzero immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h != 0 || data.l != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSNP(tetra instr) {
  /* conditionally set if nonpositive */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSNPI(tetra instr) {
  /* conditionally set if nonpositive immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


static void CSEV(tetra instr) {
  /* conditionally set if even */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) == 0) {
    data = readDR(z);
    writeDR(x, data);
  }
}


static void CSEVI(tetra instr) {
  /* conditionally set if even immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
    writeDR(x, data);
  }
}


/*------------------------------------------------------------*/

/* 70 - 7F */


static void ZSN(tetra instr) {
  /* zero or set if negative */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNI(tetra instr) {
  /* zero or set if negative immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSZ(tetra instr) {
  /* zero or set if zero */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h == 0 && data.l == 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSZI(tetra instr) {
  /* zero or set if zero immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h == 0 && data.l == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSP(tetra instr) {
  /* zero or set if positive */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSPI(tetra instr) {
  /* zero or set if positive immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0 && (data.h != 0 || data.l != 0)) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSOD(tetra instr) {
  /* zero or set if odd */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) != 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSODI(tetra instr) {
  /* zero or set if odd immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNN(tetra instr) {
  /* zero or set if nonnegative */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNNI(tetra instr) {
  /* zero or set if nonnegative immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNZ(tetra instr) {
  /* zero or set if nonzero */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h != 0 || data.l != 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNZI(tetra instr) {
  /* zero or set if nonzero immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if (data.h != 0 || data.l != 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNP(tetra instr) {
  /* zero or set if nonpositive */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSNPI(tetra instr) {
  /* zero or set if nonpositive immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.h & MSB(32)) != 0 || (data.h == 0 && data.l == 0)) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSEV(tetra instr) {
  /* zero or set if even */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) == 0) {
    data = readDR(z);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


static void ZSEVI(tetra instr) {
  /* zero or set if even immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  data = readDR(y);
  if ((data.l & 1) == 0) {
    data.h = 0;
    data.l = ZEXT(z, 8);
  } else {
    data = zero;
  }
  writeDR(x, data);
}


/*------------------------------------------------------------*/

/* 80 - 8F */


static void LDB(tetra instr) {
  /* load byte */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead8(addr);
  data.l = SEXT(data.l, 8);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDBI(tetra instr) {
  /* load byte immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead8(addr);
  data.l = SEXT(data.l, 8);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDBU(tetra instr) {
  /* load byte unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead8(addr);
  data.l = ZEXT(data.l, 8);
  data.h = 0;
  writeDR(x, data);
}


static void LDBUI(tetra instr) {
  /* load byte unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead8(addr);
  data.l = ZEXT(data.l, 8);
  data.h = 0;
  writeDR(x, data);
}


static void LDW(tetra instr) {
  /* load wyde */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead16(addr);
  data.l = SEXT(data.l, 16);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDWI(tetra instr) {
  /* load wyde immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead16(addr);
  data.l = SEXT(data.l, 16);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDWU(tetra instr) {
  /* load wyde unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead16(addr);
  data.l = ZEXT(data.l, 16);
  data.h = 0;
  writeDR(x, data);
}


static void LDWUI(tetra instr) {
  /* load wyde unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead16(addr);
  data.l = ZEXT(data.l, 16);
  data.h = 0;
  writeDR(x, data);
}


static void LDT(tetra instr) {
  /* load tetra */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead32(addr);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDTI(tetra instr) {
  /* load tetra immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead32(addr);
  if (data.l & MSB(32)) {
    data.h = -1;
  } else {
    data.h = 0;
  }
  writeDR(x, data);
}


static void LDTU(tetra instr) {
  /* load tetra unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.l = mmuRead32(addr);
  data.h = 0;
  writeDR(x, data);
}


static void LDTUI(tetra instr) {
  /* load tetra unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.l = mmuRead32(addr);
  data.h = 0;
  writeDR(x, data);
}


static void LDO(tetra instr) {
  /* load octa */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


static void LDOI(tetra instr) {
  /* load octa immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


static void LDOU(tetra instr) {
  /* load octa unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


static void LDOUI(tetra instr) {
  /* load octa unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


/*------------------------------------------------------------*/

/* 90 - 9F */


static void LDHT(tetra instr) {
  /* load high tetra */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.h = mmuRead32(addr);
  data.l = 0;
  writeDR(x, data);
}


static void LDHTI(tetra instr) {
  /* load high tetra immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.h = mmuRead32(addr);
  data.l = 0;
  writeDR(x, data);
}


static void CSWAP(tetra instr) {
  /* compare and swap octabytes */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  if (data.h == gr[rP].h && data.l == gr[rP].l) {
    data = readDR(x);
    mmuWrite64(addr, data);
    data = one;
  } else {
    gr[rP] = data;
    data = zero;
  }
  writeDR(x, data);
}


static void CSWAPI(tetra instr) {
  /* compare and swap octabytes immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  if (data.h == gr[rP].h && data.l == gr[rP].l) {
    data = readDR(x);
    mmuWrite64(addr, data);
    data = one;
  } else {
    gr[rP] = data;
    data = zero;
  }
  writeDR(x, data);
}


static void LDUNC(tetra instr) {
  /* load octa uncached */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


static void LDUNCI(tetra instr) {
  /* load octa uncached immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = mmuRead64(addr);
  writeDR(x, data);
}


static void PRELD(tetra instr) {
  /* preload data */
}


static void PRELDI(tetra instr) {
  /* preload data immediate */
}


static void PREGO(tetra instr) {
  /* prefetch to go */
}


static void PREGOI(tetra instr) {
  /* prefetch to go immediate */
}


static void GO(tetra instr) {
  /* go to location */
  int x, y, z;
  octa op1, op2;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  writeDR(x, pc);
  pc = addOcta(op1, op2);
}


static void GOI(tetra instr) {
  /* go to location immediate */
  int x, y, z;
  octa op1, op2;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  writeDR(x, pc);
  pc = addOcta(op1, op2);
}


/*------------------------------------------------------------*/

/* A0 - AF */


static void STB(tetra instr) {
  /* store byte */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 56), 56);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite8(addr, (byte) data.l);
}


static void STBI(tetra instr) {
  /* store byte immediate */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 56), 56);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite8(addr, (byte) data.l);
}


static void STBU(tetra instr) {
  /* store byte unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite8(addr, (byte) data.l);
}


static void STBUI(tetra instr) {
  /* store byte unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite8(addr, (byte) data.l);
}


static void STW(tetra instr) {
  /* store wyde */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 48), 48);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite16(addr, (wyde) data.l);
}


static void STWI(tetra instr) {
  /* store wyde immediate */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 48), 48);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite16(addr, (wyde) data.l);
}


static void STWU(tetra instr) {
  /* store wyde unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite16(addr, (wyde) data.l);
}


static void STWUI(tetra instr) {
  /* store wyde unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite16(addr, (wyde) data.l);
}


static void STT(tetra instr) {
  /* store tetra */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 32), 32);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite32(addr, data.l);
}


static void STTI(tetra instr) {
  /* store tetra immediate */
  int x, y, z;
  octa op1, op2, addr, data;
  octa temp;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  temp = sarOcta(shlOcta(data, 32), 32);
  if (temp.h != data.h || temp.l != data.l) {
    /* !!!!! signal integer overflow exception !!!!! */
  }
  mmuWrite32(addr, data.l);
}


static void STTU(tetra instr) {
  /* store tetra unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite32(addr, data.l);
}


static void STTUI(tetra instr) {
  /* store tetra unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite32(addr, data.l);
}


static void STO(tetra instr) {
  /* store octa */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


static void STOI(tetra instr) {
  /* store octa immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


static void STOU(tetra instr) {
  /* store octa unsigned */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


static void STOUI(tetra instr) {
  /* store octa unsigned immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


/*------------------------------------------------------------*/

/* B0 - BF */


static void STHT(tetra instr) {
  /* store high tetra */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite32(addr, data.h);
}


static void STHTI(tetra instr) {
  /* store high tetra immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite32(addr, data.h);
}


static void STCO(tetra instr) {
  /* store constant octabyte */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data.h = 0;
  data.l = ZEXT(x, 8);
  mmuWrite64(addr, data);
}


static void STCOI(tetra instr) {
  /* store constant octabyte immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data.h = 0;
  data.l = ZEXT(x, 8);
  mmuWrite64(addr, data);
}


static void STUNC(tetra instr) {
  /* store octa uncached */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


static void STUNCI(tetra instr) {
  /* store octa uncached immediate */
  int x, y, z;
  octa op1, op2, addr, data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  addr = addOcta(op1, op2);
  data = readDR(x);
  mmuWrite64(addr, data);
}


static void PREST(tetra instr) {
  /* prestore data */
}


static void PRESTI(tetra instr) {
  /* prestore data immediate */
}


static void PUSHGO(tetra instr) {
  /* push registers and go */
  int x, y, z;
  octa op1, op2;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  push(x);
  gr[rJ] = pc;
  pc = addOcta(op1, op2);
}


static void PUSHGOI(tetra instr) {
  /* push registers and go immediate */
  int x, y, z;
  octa op1, op2;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  push(x);
  gr[rJ] = pc;
  pc = addOcta(op1, op2);
}


/*------------------------------------------------------------*/

/* C0 - CF */


static void OR(tetra instr) {
  /* bitwise or */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = orOcta(op1, op2);
  writeDR(x, res);
}


static void ORI(tetra instr) {
  /* bitwise or immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = orOcta(op1, op2);
  writeDR(x, res);
}


static void ORN(tetra instr) {
  /* bitwise or-not */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = ornOcta(op1, op2);
  writeDR(x, res);
}


static void ORNI(tetra instr) {
  /* bitwise or-not immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = ornOcta(op1, op2);
  writeDR(x, res);
}


static void NOR(tetra instr) {
  /* bitwise not-or */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = norOcta(op1, op2);
  writeDR(x, res);
}


static void NORI(tetra instr) {
  /* bitwise not-or immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = norOcta(op1, op2);
  writeDR(x, res);
}


static void XOR(tetra instr) {
  /* bitwise exclusive-or */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = xorOcta(op1, op2);
  writeDR(x, res);
}


static void XORI(tetra instr) {
  /* bitwise exclusive-or immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = xorOcta(op1, op2);
  writeDR(x, res);
}


static void AND(tetra instr) {
  /* bitwise and */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = andOcta(op1, op2);
  writeDR(x, res);
}


static void ANDI(tetra instr) {
  /* bitwise and immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = andOcta(op1, op2);
  writeDR(x, res);
}


static void ANDN(tetra instr) {
  /* bitwise and-not */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = andnOcta(op1, op2);
  writeDR(x, res);
}


static void ANDNI(tetra instr) {
  /* bitwise and-not immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = andnOcta(op1, op2);
  writeDR(x, res);
}


static void NAND(tetra instr) {
  /* bitwise not-and */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = nandOcta(op1, op2);
  writeDR(x, res);
}


static void NANDI(tetra instr) {
  /* bitwise not-and immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = nandOcta(op1, op2);
  writeDR(x, res);
}


static void NXOR(tetra instr) {
  /* bitwise not-exclusive-or */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2 = readDR(z);
  res = nxorOcta(op1, op2);
  writeDR(x, res);
}


static void NXORI(tetra instr) {
  /* bitwise not-exclusive-or immediate */
  int x, y, z;
  octa op1, op2, res;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  op1 = readDR(y);
  op2.h = 0;
  op2.l = ZEXT(z, 8);
  res = nxorOcta(op1, op2);
  writeDR(x, res);
}


/*------------------------------------------------------------*/

/* D0 - DF */


/*------------------------------------------------------------*/

/* E0 - EF */


static void SETH(tetra instr) {
  /* set to high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz << 16;
  imm.l = 0;
  res = imm;
  writeDR(x, res);
}


static void SETMH(tetra instr) {
  /* set to medium high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz;
  imm.l = 0;
  res = imm;
  writeDR(x, res);
}


static void SETML(tetra instr) {
  /* set to medium low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz << 16;
  res = imm;
  writeDR(x, res);
}


static void SETL(tetra instr) {
  /* set to low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz;
  res = imm;
  writeDR(x, res);
}


static void INCH(tetra instr) {
  /* increase by high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz << 16;
  imm.l = 0;
  res = readDR(x);
  res = addOcta(res, imm);
  writeDR(x, res);
}


static void INCMH(tetra instr) {
  /* increase by medium high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz;
  imm.l = 0;
  res = readDR(x);
  res = addOcta(res, imm);
  writeDR(x, res);
}


static void INCML(tetra instr) {
  /* increase by medium low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz << 16;
  res = readDR(x);
  res = addOcta(res, imm);
  writeDR(x, res);
}


static void INCL(tetra instr) {
  /* increase by low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz;
  res = readDR(x);
  res = addOcta(res, imm);
  writeDR(x, res);
}


static void ORH(tetra instr) {
  /* bitwise or with high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz << 16;
  imm.l = 0;
  res = readDR(x);
  res = orOcta(res, imm);
  writeDR(x, res);
}


static void ORMH(tetra instr) {
  /* bitwise or with medium high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz;
  imm.l = 0;
  res = readDR(x);
  res = orOcta(res, imm);
  writeDR(x, res);
}


static void ORML(tetra instr) {
  /* bitwise or with medium low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz << 16;
  res = readDR(x);
  res = orOcta(res, imm);
  writeDR(x, res);
}


static void ORL(tetra instr) {
  /* bitwise or with low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz;
  res = readDR(x);
  res = orOcta(res, imm);
  writeDR(x, res);
}


static void ANDNH(tetra instr) {
  /* bitwise and-not high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz << 16;
  imm.l = 0;
  res = readDR(x);
  res = andnOcta(res, imm);
  writeDR(x, res);
}


static void ANDNMH(tetra instr) {
  /* bitwise and-not medium high wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = yz;
  imm.l = 0;
  res = readDR(x);
  res = andnOcta(res, imm);
  writeDR(x, res);
}


static void ANDNML(tetra instr) {
  /* bitwise and-not medium low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz << 16;
  res = readDR(x);
  res = andnOcta(res, imm);
  writeDR(x, res);
}


static void ANDNL(tetra instr) {
  /* bitwise and-not low wyde */
  int x, yz;
  octa imm, res;

  x = DST(instr);
  yz = LIT16(instr);
  imm.h = 0;
  imm.l = yz;
  res = readDR(x);
  res = andnOcta(res, imm);
  writeDR(x, res);
}


/*------------------------------------------------------------*/

/* F0 - FF */


static void JMP(tetra instr) {
  /* jump */
  int offset;

  offset = LIT24(instr) << 2;
  pc = incrOcta(pc, offset - 4);
}


static void JMPB(tetra instr) {
  /* jump back */
  int offset;

  offset = (LIT24(instr) - 0x1000000) << 2;
  pc = incrOcta(pc, offset - 4);
}


static void PUSHJ(tetra instr) {
  /* push registers and jump */
  int x;
  int offset;

  x = DST(instr);
  offset = LIT16(instr) << 2;
  push(x);
  gr[rJ] = pc;
  pc = incrOcta(pc, offset - 4);
}


static void PUSHJB(tetra instr) {
  /* push registers and jump back */
  int x;
  int offset;

  x = DST(instr);
  offset = (LIT16(instr) - 0x10000) << 2;
  push(x);
  gr[rJ] = pc;
  pc = incrOcta(pc, offset - 4);
}


static void GETA(tetra instr) {
  /* get address */
  int x;
  int offset;
  octa addr;

  x = DST(instr);
  offset = LIT16(instr) << 2;
  addr = incrOcta(pc, offset - 4);
  writeDR(x, addr);
}


static void GETAB(tetra instr) {
  /* get address back */
  int x;
  int offset;
  octa addr;

  x = DST(instr);
  offset = (LIT16(instr) - 0x10000) << 2;
  addr = incrOcta(pc, offset - 4);
  writeDR(x, addr);
}


static void PUT(tetra instr) {
  /* put into special register */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  if (y != 0 || x >= 32) {
    /* !!!!! signal illegal instruction exception !!!!! */
  }
  data = readDR(z);
  writeSpecial(x, data);
}


static void PUTI(tetra instr) {
  /* put into special register immediate */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  if (y != 0 || x >= 32) {
    /* !!!!! signal illegal instruction exception !!!!! */
  }
  data.h = 0;
  data.l = ZEXT(z, 8);
  writeSpecial(x, data);
}


static void POP(tetra instr) {
  /* pop registers and return from subroutine */
  int x;
  int offset;

  x = DST(instr);
  offset = LIT16(instr) << 2;
  pop(x);
  pc = incrOcta(gr[rJ], offset);
}


static void SWYM(tetra instr) {
  /* sympathize with your machinery */
}


static void GET(tetra instr) {
  /* get from special register */
  int x, y, z;
  octa data;

  x = DST(instr);
  y = SRC1(instr);
  z = SRC2(instr);
  if (y != 0 || z >= 32) {
    /* !!!!! signal illegal instruction exception !!!!! */
  }
  data = readSpecial(z);
  writeDR(x, data);
}


static void TRIP(tetra instr) {
  /* !!!!! not implemented yet !!!!! */
  ill(instr);
}


/*------------------------------------------------------------*/


static void (*instrPtr[256])(tetra instr) = {
  /* 00 */  TRAP,   ill,    ill,    ill,    ill,    ill,    ill,    ill,
  /* 08 */  ill,    ill,    ill,    ill,    ill,    ill,    ill,    ill,
  /* 10 */  ill,    ill,    ill,    ill,    ill,    ill,    ill,    ill,
  /* 18 */  MUL,    MULI,   MULU,   MULUI,  DIV,    DIVI,   DIVU,   DIVUI,
  /* 20 */  ADD,    ADDI,   ADDU,   ADDUI,  SUB,    SUBI,   SUBU,   SUBUI,
  /* 28 */  ADDU2,  ADDUI2, ADDU4,  ADDUI4, ADDU8,  ADDUI8, ADDU16, ADDUI16,
  /* 30 */  CMP,    CMPI,   CMPU,   CMPUI,  NEG,    NEGI,   NEGU,   NEGUI,
  /* 38 */  SL,     SLI,    SLU,    SLUI,   SR,     SRI,    SRU,    SRUI,
  /* 40 */  BN,     BNB,    BZ,     BZB,    BP,     BPB,    BOD,    BODB,
  /* 48 */  BNN,    BNNB,   BNZ,    BNZB,   BNP,    BNPB,   BEV,    BEVB,
  /* 50 */  PBN,    PBNB,   PBZ,    PBZB,   PBP,    PBPB,   PBOD,   PBODB,
  /* 58 */  PBNN,   PBNNB,  PBNZ,   PBNZB,  PBNP,   PBNPB,  PBEV,   PBEVB,
  /* 60 */  CSN,    CSNI,   CSZ,    CSZI,   CSP,    CSPI,   CSOD,   CSODI,
  /* 68 */  CSNN,   CSNNI,  CSNZ,   CSNZI,  CSNP,   CSNPI,  CSEV,   CSEVI,
  /* 70 */  ZSN,    ZSNI,   ZSZ,    ZSZI,   ZSP,    ZSPI,   ZSOD,   ZSODI,
  /* 78 */  ZSNN,   ZSNNI,  ZSNZ,   ZSNZI,  ZSNP,   ZSNPI,  ZSEV,   ZSEVI,
  /* 80 */  LDB,    LDBI,   LDBU,   LDBUI,  LDW,    LDWI,   LDWU,   LDWUI,
  /* 88 */  LDT,    LDTI,   LDTU,   LDTUI,  LDO,    LDOI,   LDOU,   LDOUI,
  /* 90 */  ill,    ill,    LDHT,   LDHTI,  CSWAP,  CSWAPI, LDUNC,  LDUNCI,
  /* 98 */  ill,    ill,    PRELD,  PRELDI, PREGO,  PREGOI, GO,     GOI,
  /* A0 */  STB,    STBI,   STBU,   STBUI,  STW,    STWI,   STWU,   STWUI,
  /* A8 */  STT,    STTI,   STTU,   STTUI,  STO,    STOI,   STOU,   STOUI,
  /* B0 */  ill,    ill,    STHT,   STHTI,  STCO,   STCOI,  STUNC,  STUNCI,
  /* B8 */  ill,    ill,    PREST,  PRESTI, ill,    ill,    PUSHGO, PUSHGOI,
  /* C0 */  OR,     ORI,    ORN,    ORNI,   NOR,    NORI,   XOR,    XORI,
  /* C8 */  AND,    ANDI,   ANDN,   ANDNI,  NAND,   NANDI,  NXOR,   NXORI,
  /* D0 */  ill,    ill,    ill,    ill,    ill,    ill,    ill,    ill,
  /* D8 */  ill,    ill,    ill,    ill,    ill,    ill,    ill,    ill,
  /* E0 */  SETH,   SETMH,  SETML,  SETL,   INCH,   INCMH,  INCML,  INCL,
  /* E8 */  ORH,    ORMH,   ORML,   ORL,    ANDNH,  ANDNMH, ANDNML, ANDNL,
  /* F0 */  JMP,    JMPB,   PUSHJ,  PUSHJB, GETA,   GETAB,  PUT,    PUTI,
  /* F8 */  POP,    ill,    ill,    ill,    ill,    SWYM,   GET,    TRIP
};


static void execNextInstr(void) {
  tetra instr;

  /* fetch instruction */
  instr = mmuRead32(pc);
  pc = incrOcta(pc, 4);
  /* execute instruction */
  (*instrPtr[OPCODE(instr)])(instr);
  /* handle "real-time" events */
  if (++instrCount == INSTRS_PER_MSEC) {
    instrCount = 0;
    timerTick();
  }
}


/*------------------------------------------------------------*/


octa cpuGetPC(void) {
  return pc;
}


void cpuSetPC(octa address) {
  address.l &= ~0x00000003;
  pc = address;
}


octa cpuGetGR(int regnum) {
  regnum &= GR_MASK;
  if (regnum < 32) {
    return readSpecial(regnum);
  } else {
    return gr[regnum];
  }
}


void cpuSetGR(int regnum, octa value) {
  regnum &= GR_MASK;
  if (regnum < 32) {
    writeSpecial(regnum, value);
  } else {
    gr[regnum] = value;
  }
}


octa cpuGetLR(int regnum) {
  regnum &= LR_MASK;
  return lr[regnum];
}


void cpuSetLR(int regnum, octa value) {
  regnum &= LR_MASK;
  lr[regnum] = value;
}


octa cpuGetDR(int regnum) {
  regnum &= 0xFF;
  return readDR(regnum);
}


void cpuSetDR(int regnum, octa value) {
  regnum &= 0xFF;
  writeDR(regnum, value);
}


bool cpuTestBreak(void) {
  return breakSet;
}


octa cpuGetBreak(void) {
  return breakAddr;
}


void cpuSetBreak(octa address) {
  address.l &= ~0x00000003;
  breakAddr = address;
  breakSet = TRUE;
}


void cpuResetBreak(void) {
  breakSet = FALSE;
}


void cpuStep(void) {
  jmp_buf myEnvironment;
  int exception;

  exception = setjmp(myEnvironment);
  if (exception == NO_EXCEPTION) {
    /* first call */
    pushEnvironment(&myEnvironment);
    execNextInstr();
  } else {
    /* exception thrown */
    handleRuntimeException(exception);
  }
  popEnvironment();
}


void cpuContinue(void) {
  jmp_buf myEnvironment;
  int exception;

  exception = setjmp(myEnvironment);
  if (exception == NO_EXCEPTION) {
    /* first call */
    pushEnvironment(&myEnvironment);
  } else {
    /* exception thrown */
    handleRuntimeException(exception);
  }
  while (1) {
    execNextInstr();
    if (breakSet && pc.h == breakAddr.h && pc.l == breakAddr.l) {
      break;
    }
  }
  popEnvironment();
}


void cpuReset(void) {
  int i;

  cPrintf("Resetting CPU...\n");
  /* preset registers */
  pc.h = ROM_BASE_H | MSB(32);
  pc.l = ROM_BASE_L;
  for (i = 0; i < GR_NUM; i++) {
    gr[i] = zero;
  }
  for (i = 0; i < LR_NUM; i++) {
    lr[i] = zero;
  }
  G = 255;
  gr[rG].h = 0;
  gr[rG].l = G;
  L = 0;
  gr[rL].h = 0;
  gr[rL].l = L;
  O = 0;
  gr[rO].h = 0;
  gr[rO].l = O;
  S = 0;
  gr[rS].h = 0;
  gr[rS].l = S;
  gr[rN].h = (VERSION << 24) | (SUBVERSION << 16) | (SUBSUBVERSION << 8);
  gr[rN].l = ABSTIME;
  /* reset simulator control variables */
  breakSet = FALSE;
  breakAddr = zero;
  instrCount = 0;
}


void cpuInit(void) {
  cpuReset();
}


void cpuExit(void) {
}
