/*
 * vat.c -- virtual address translation for GIMMIX
 */


#include <stdio.h>
#include <stdlib.h>
#include <string.h>


#define LINE_SIZE	100


typedef unsigned int tetra;

typedef struct {
  tetra h;
  tetra l;
} octa;


octa one = {
  0, 1
};

octa rV;
unsigned int b[5];
unsigned int s;
unsigned int r;
unsigned int n;
unsigned int f;

octa pte;
octa ptp[4];  /* PTP at level i is stored in ptp[i - 1], i = 1..4 */


octa incr(octa y, int delta) {
  octa x;

  x.h = y.h;
  x.l = y.l + delta;
  if (delta >= 0) {
    if (x.l < y.l) {
      x.h++;
    }
  } else {
    if (x.l > y.l) {
      x.h--;
    }
  }
  return x;
}


octa and(octa y, octa z) {
  octa x;

  x.h = y.h & z.h;
  x.l = y.l & z.l;
  return x;
}


octa or(octa y, octa z) {
  octa x;

  x.h = y.h | z.h;
  x.l = y.l | z.l;
  return x;
}


octa not(octa y) {
  octa x;

  x.h = ~y.h;
  x.l = ~y.l;
  return x;
}


octa shiftLeft(octa y, int cnt) {
  octa x;

  x = y;
  while (cnt--) {
    x.h <<= 1;
    if (x.l & 0x80000000) {
      x.h |= 0x00000001;
    }
    x.l <<= 1;
  }
  return x;
}


octa shiftRight(octa y, int cnt) {
  octa x;

  x = y;
  while (cnt--) {
    x.l >>= 1;
    if (x.h & 0x00000001) {
      x.l |= 0x80000000;
    }
    x.h >>= 1;
  }
  return x;
}


void printOcta(octa o) {
  printf("0x%08x%08x", o.h, o.l);
}


octa readOcta(char **pp) {
  char *p;
  octa v;
  int c;

  p = *pp;
  v.h = 0;
  v.l = 0;
  while ((*p >= '0' && *p <= '9') ||
         (*p >= 'a' && *p <= 'f') ||
         (*p >= 'A' && *p <= 'F')) {
    if (*p >= '0' && *p <= '9') {
      c = *p - '0';
    } else
    if (*p >= 'a' && *p <= 'f') {
      c = *p - 'a' + 10;
    } else
    if (*p >= 'A' && *p <= 'F') {
      c = *p - 'A' + 10;
    }
    p++;
    v = shiftLeft(v, 4);
    v.l |= c;
  }
  *pp = p;
  return v;
}


char *skipBlanks(char *p) {
  while (*p == ' ' || *p == '\t') {
    p++;
  }
  return p;
}


void showSegmentInfo(int segment) {
  printf("b[%d] = 0x%x, b[%d] = 0x%x: ",
         segment, b[segment], segment + 1, b[segment + 1]);
  if (b[segment] > b[segment + 1]) {
    printf("Segment %d must be empty\n", segment);
  } else
  if (b[segment] == b[segment + 1]) {
    printf("Segment %d has at most 1 page\n", segment);
  } else {
    printf("Segment %d has at most 2^%d pages\n",
           segment, 10 * (b[segment + 1] - b[segment]));
  }
}


void unpack_rV(void) {
  b[0] = 0;
  b[1] = (rV.h >> 28) & 0xf;
  b[2] = (rV.h >> 24) & 0xf;
  b[3] = (rV.h >> 20) & 0xf;
  b[4] = (rV.h >> 16) & 0xf;
  s = (rV.h >> 8) & 0xff;
  r = ((rV.h & 0xff) << 19) | ((rV.l >> 13) & 0x7ffff);
  n = (rV.l >> 3) & 0x3ff;
  f = rV.l & 0x7;
}


int main(void) {
  char line[LINE_SIZE];
  char *p;
  octa o;
  int level;
  int segment;
  octa mask;
  octa base;
  octa page, offset;
  tetra a[5];
  octa address;

  printf("\n");
  printf("Virtual Address Translation for MMIX\n");
  rV.h = 0x369c2004;
  rV.l = 0x00000000;
  unpack_rV();
  pte.h = 0x00001234;
  pte.l = 0x00000007;
  for (level = 1; level <= 4; level++) {
    ptp[level - 1].h = 0x80000000 | (0x1234 << (level - 1));
    ptp[level - 1].l = 0x56780000 >> (level - 1);
  }
  while (1) {
    printf("\n");
    printf("Please enter command:\n");
    printf("  s          show relevant translation values\n");
    printf("  v <n>      set register rV to value n (hex)\n");
    printf("  e <n>      set PTE to value n (hex)\n");
    printf("  p <i> <n>  set PTP at level i to value n (hex)\n");
    printf("  xv         explain the current rV value\n");
    printf("  xe         explain the current PTE value\n");
    printf("  xp <i>     explain the current PTP value at level i\n");
    printf("  t <n>      translate virtual address n (hex)\n");
    printf("  q          quit\n");
    printf("\n? ");
    fflush(stdout);
    if (fgets(line, LINE_SIZE, stdin) == NULL) {
      printf("\n");
      break;
    }
    p = skipBlanks(line);
    if (*p == '\n') {
      continue;
    }
    if (*p == 's') {
      p = skipBlanks(p + 1);
      if (*p != '\n') {
        printf("Warning: garbage following 's' command!\n");
      }
      printf("\n");
      printf("rV\t\t");
      printOcta(rV);
      printf("\n");
      printf("PTE\t\t");
      printOcta(pte);
      printf("\n");
      for (level = 1; level <= 4; level++) {
        printf("PTP at level %d\t", level);
        printOcta(ptp[level - 1]);
        printf("\n");
      }
      continue;
    }
    if (*p == 'v') {
      p = skipBlanks(p + 1);
      if (*p == '\n') {
        printf("Error: number required for 'v' command!\n");
        continue;
      }
      o = readOcta(&p);
      if (*p != '\n') {
        printf("Warning: garbage following 'v' command!\n");
      }
      rV = o;
      unpack_rV();
      continue;
    }
    if (*p == 'e') {
      p = skipBlanks(p + 1);
      if (*p == '\n') {
        printf("Error: number required for 'e' command!\n");
        continue;
      }
      o = readOcta(&p);
      if (*p != '\n') {
        printf("Warning: garbage following 'e' command!\n");
      }
      pte = o;
      continue;
    }
    if (*p == 'p') {
      p = skipBlanks(p + 1);
      if (*p == '\n') {
        printf("Error: number required for 'p' command!\n");
        continue;
      }
      o = readOcta(&p);
      if (o.h != 0 || o.l < 1 || o.l > 4) {
        printf("Error: PTP level must be in range 1..4!\n");
        continue;
      }
      level = o.l;
      p = skipBlanks(p);
      if (*p == '\n') {
        printf("Error: another number required for 'p' command!\n");
        continue;
      }
      o = readOcta(&p);
      if (*p != '\n') {
        printf("Warning: garbage following 'p' command!\n");
      }
      ptp[level - 1] = o;
      continue;
    }
    if (*p == 'x') {
      p++;
      if (*p == '\n') {
        printf("Error: explain what?\n");
        continue;
      }
      if (*p == 'v') {
        p = skipBlanks(p + 1);
        if (*p != '\n') {
          printf("Warning: garbage following 'xv' command!\n");
        }
        printf("\n");
        printf("The current value of rV is ");
        printOcta(rV);
        printf(":\n");
        for (segment = 0; segment < 4; segment++) {
          showSegmentInfo(segment);
        }
        printf("Page size is 2^%d bytes", s);
        if (s < 13 || s > 48) {
          printf(": THIS IS ILLEGAL!\n");
        } else {
          printf("\n");
        }
        printf("Root location is 0x%08x\n", r);
        printf("Address space number is %d\n", n);
        printf("Function field is %d", f);
        if (f == 0) {
          printf(" (virtual address translation done by hardware)\n");
        } else
        if (f == 1) {
          printf(" (virtual address translation done by software)\n");
        } else {
          printf(": THIS IS ILLEGAL!\n");
        }
        continue;
      }
      if (*p == 'e') {
        p = skipBlanks(p + 1);
        if (*p != '\n') {
          printf("Warning: garbage following 'xe' command!\n");
        }
        printf("\n");
        printf("The current value of the PTE is ");
        printOcta(pte);
        printf(":\n");
        mask = incr(shiftLeft(one, s), -1);
        mask.h |= 0xffff0000;
        mask = not(mask);
        base = and(pte, mask);
        printf("Physical page base address is ");
        printOcta(base);
        printf("\n");
        printf("Address space number is %d\n", (pte.l >> 3) & 0x3ff);
        printf("Protection bits are '%c%c%c'\n",
               (pte.l & 4) ? 'r' : '-',
               (pte.l & 2) ? 'w' : '-',
               (pte.l & 1) ? 'x' : '-');
        continue;
      }
      if (*p == 'p') {
        p = skipBlanks(p + 1);
        if (*p == '\n') {
          printf("Error: number required for 'xp' command!\n");
          continue;
        }
        o = readOcta(&p);
        if (o.h != 0 || o.l < 1 || o.l > 4) {
          printf("Error: PTP level must be in range 1..4!\n");
          continue;
        }
        level = o.l;
        if (*p != '\n') {
          printf("Warning: garbage following 'xp' command!\n");
        }
        printf("\n");
        printf("The current value of the PTP at level %d is ", level);
        printOcta(ptp[level - 1]);
        printf(":\n");
        if ((ptp[level - 1].h & 0x80000000) == 0) {
          printf("MSB of PTP not set: THIS IS ILLEGAL!\n");
        }
        mask = incr(shiftLeft(one, 13), -1);
        mask.h |= 0x80000000;
        mask = not(mask);
        base = and(ptp[level - 1], mask);
        printf("Physical page table base address of level %d is ", level - 1);
        printOcta(base);
        printf("\n");
        printf("Address space number is %d\n",
               (ptp[level - 1].l >> 3) & 0x3ff);
        printf("Protection bits are '%c%c%c' (ignored)\n",
               (ptp[level - 1].l & 4) ? 'r' : '-',
               (ptp[level - 1].l & 2) ? 'w' : '-',
               (ptp[level - 1].l & 1) ? 'x' : '-');
        continue;
      }
      printf("Error: cannot understand this 'x' command!\n");
      continue;
    }
    if (*p == 't') {
      p = skipBlanks(p + 1);
      if (*p == '\n') {
        printf("Error: number required for 't' command!\n");
        continue;
      }
      o = readOcta(&p);
      if (*p != '\n') {
        printf("Warning: garbage following 't' command!\n");
      }
      printf("\n");
      printf("The virtual address ");
      printOcta(o);
      printf(" is translated as follows:\n");
      if (o.h & 0x80000000) {
        printf("MSB of virtual address set: address is physical address\n");
        printf("ACCESS IS PRIVILEGED\n");
        o.h &= 0x7fffffff;
        printf("Physical address = ");
        printOcta(o);
        printf("\n");
        if (o.h & 0x7fff0000) {
          printf("This physical address accesses an I/O device\n");
        } else {
          printf("This physical address accesses main memory\n");
        }
      } else {
        segment = (o.h >> 29) & 0x3;
        printf("The segment number is %d\n", segment);
        o.h &= 0x1fffffff;
        printf("The segment-relative virtual address is ");
        printOcta(o);
        printf("\n");
        mask = incr(shiftLeft(one, s), -1);
        printf("The page mask is ");
        printOcta(mask);
        printf("\n");
        page = shiftRight(o, s);
        offset = and(o, mask);
        printf("This leads to page ");
        printOcta(page);
        printf(", offset ");
        printOcta(offset);
        printf("\n");
        for (level = 0; level < 5; level++) {
          a[level] = shiftRight(page, 10 * level).l & 0x3ff;
        }
        printf("Thus the page number in radix 1024 is ");
        printf("(0x%03x 0x%03x 0x%03x 0x%03x 0x%03x)\n",
               a[4], a[3], a[2], a[1], a[0]);
        printf("Translation chain: ");
        if (a[4] != 0) {
          printf("VA --> PTP(4) --> PTP(3) --> PTP(2) --> PTP(1) --> PTE\n");
          level = 4;
        } else
        if (a[3] != 0) {
          printf("VA --> PTP(3) --> PTP(2) --> PTP(1) --> PTE\n");
          level = 3;
        } else
        if (a[2] != 0) {
          printf("VA --> PTP(2) --> PTP(1) --> PTE\n");
          level = 2;
        } else
        if (a[1] != 0) {
          printf("VA --> PTP(1) --> PTE\n");
          level = 1;
        } else {
          printf("VA --> PTE\n");
          level = 0;
        }
        /* do first stage of lookup */
        base.h = 0;
        base.l = r;
        base = incr(base, b[segment] + level);
        base = shiftLeft(base, 13);
        address = incr(base, a[level] << 3);
        printf("Looking up memory at address\n");
        printf("  ((0x%08x + 0x%02x + %d) << 13) + (0x%03x << 3) = ",
               r, b[segment], level, a[level]);
        printOcta(address);
        printf("\n");
        /* interpret auxiliary PTP values */
        for (; level > 0; level--) {
          printf("The PTP value at level %d is ", level);
          printOcta(ptp[level - 1]);
          printf("\n");
          mask = incr(shiftLeft(one, 13), -1);
          mask.h |= 0x80000000;
          mask = not(mask);
          base = and(ptp[level - 1], mask);
          address = incr(base, a[level - 1] << 3);
          printf("Looking up memory at address\n");
          printf("  ");
          printOcta(base);
          printf(" + (0x%03x << 3) = ", a[level - 1]);
          printOcta(address);
          printf("\n");
        }
        /* interpret PTE value */
        printf("The PTE value is ");
        printOcta(pte);
        printf("\n");
        mask = incr(shiftLeft(one, s), -1);
        mask.h |= 0xffff0000;
        mask = not(mask);
        base = and(pte, mask);
        address = or(base, offset);
        printf("Physical page base address is ");
        printOcta(base);
        printf("\n");
        printf("Then, finally, the physical address is ");
        printOcta(address);
        printf("\n");
      }
      continue;
    }
    if (*p == 'q') {
      p = skipBlanks(p + 1);
      if (*p != '\n') {
        printf("Warning: garbage following 'q' command!\n");
      }
      break;
    }
    printf("Error: cannot understand your command, please re-enter!\n");
  }
  printf("\nBye!\n\n");
  return 0;
}
